package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Queue;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonToken;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


/*
 * Needs to be called AFTER term normalization.
 */

/**
 * As of Oct. 23 this is not a transformation anymore. This "transformation" merely collects type definitions for later reference!
 * See commenting/uncommenting.
 *
 */
public class TypeDefinitions extends Transformation {
	@SuppressWarnings(value="unchecked")
	public TypeDefinitions() {
		super();
		
		for (String defStr : implicitTypeDefinitionStrings) {
			ANTLRStringStream stream = new ANTLRStringStream(defStr);
			DuraLexer lexer = new DuraLexer(stream);
			CommonTokenStream tokens = new CommonTokenStream(lexer);
			DuraParser parser = new DuraParser(tokens);
			
			try {
				CommonTree term = (CommonTree) parser.typeDefinition().getTree();
				
				int index = term.getChildCount() - 1;
				CommonTree maybeSupplement = (CommonTree) term.getChild(index);
				if (maybeSupplement.getType() == DuraParser.SUPPLEMENT) {
					term.deleteChild(index);
				}
				Collections.sort(term.getChildren(), termComparator);
				term.freshenParentAndChildIndexes();
				for (Object c : term.getChildren()) {
					CommonTree child = (CommonTree) c;
					
					// All children are terms at this stage.
					// if (child.getType() == DuraParser.TERM) {
					normalizeTerm(child);
					// }
				}
				if (maybeSupplement.getType() == DuraParser.SUPPLEMENT) {
					term.addChild(maybeSupplement);
				}
				
				implicitTypeDefinitions.add(term);
			} catch (RecognitionException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
	
	@Override
	@SuppressWarnings(value="unchecked")
	public void apply(CommonTree tree) {
		HashMap<String, CommonTree> nameToTypeDefMap = new HashMap<String, CommonTree>();
		
		for (CommonTree typeDef : implicitTypeDefinitions) {
			// This should not happen.
			if (nameToTypeDefMap.get(typeDef.getText()) != null) {
				throw new RuntimeException("Type name \"" + typeDef.getText() + "\" is already bound to a different definition.");
			}
			
			nameToTypeDefMap.put(typeDef.getText(), typeDef);
		}
		
		CommonTree searchRoot = (CommonTree) tree.getChild(0).getChild(0);
		if (searchRoot.getChildCount() > 0) {
			List<CommonTree> children = searchRoot.getChildren();
			for (CommonTree child : children) {
				if (child.getType() != DuraParser.TYPEDEF) {
					continue;
				}
			
				int index = child.getChildCount() - 1;
				CommonTree maybeSupplement = (CommonTree) child.getChild(index);
				if (maybeSupplement.getType() == DuraParser.SUPPLEMENT) {
					child.deleteChild(index);
				}
				Collections.sort(child.getChildren(), termComparator);
				child.freshenParentAndChildIndexes();
				if (maybeSupplement.getType() == DuraParser.SUPPLEMENT) {
					child.addChild(maybeSupplement);
				}
				
				CommonTree temp = nameToTypeDefMap.get(child.getText());
				if (temp != null) {
					// Let us allow type re-definitions (only) if they are equal.
					if (treesEqual(child, temp)) {
						continue;
					} else {
						throw new RuntimeException("Type name \"" + child.getText() + "\" is already bound to a different (possibly implicit) definition (" + temp.toStringTree() + ").");
					}
				}
			
				nameToTypeDefMap.put(child.getText(), child);
			}
			
			// Removes type definitions from syntax tree!
			// children.clear();
			// searchRoot.freshenParentAndChildIndexes();
		}
		
		// resolveInternalTypeReferences(nameToTypeDefMap);
		// resolveExternalTypeReferences(tree, nameToTypeDefMap);
		
		typeDefinitionsMap = nameToTypeDefMap;
	}
	
	
	/**
	 * 
	 *  Why would we need this?
	 *  
	 *  Originally because DuraC did not support composite types!
	 *  
	 *  Now...
	 *  
	 *  
	 * 
	 *  One important aspect of some of the desugaring transformations are extensions of schemata.
	 * 
	 *  Let's say we have a schema x{ t{ int } } and we want to add id{ identifier }.
	 * 
	 *  Easy: x{ id{ identifier }, t{ int } }
	 * 
	 * 
	 *  Now consider a complex composite type called "blub" consisting of a list.
	 * 
	 *  blub = [ g{ float }, l{ int } ]
	 * 
	 *  ...and a schema x{ blub }, which is allowed in the grammars of DuraC and Dura
	 *
	 *  How can we extend this? Not directly.  x{ blub, id{ identifier } } is NOT permitted by the grammar of DuraC.
	 *  
	 *  We CAN however first resolve the complex composite type blub so that the schema becomes:
	 *  
	 *  x{ g{ float }, l{ int } } 
	 *  
	 *  ...and subsequently extend this schema:
	 *  
	 *  x{ g{ float }, l{ int }, id{ identifier } } 
	 *  
	 *  
	 *  !!!
	 *  NOTE: Just because the grammar allows the initial problematic situation, it doesn't have to mean it is allowed in DuraC...
	 *  !!!
	 *  
	 *  Also similar to SchemaMatching... we use this as a workaround for the fact that no proper semantic analysis
	 *  is available/ready yet.
	 *  
	 *  Ex. Schema: t{ blub }
	 *  
	 *      Then somewhere a query:  t{ g{ var X } }
	 *      
	 *      What is the type of var X?  For that we would need to traverse up the chain of possibly chained complex composite types
	 *      until we find one that contains the term labeled "g"...
	 *      
	 *  If we replace the complex type properly (in general this might include constraints) we can expand like this:
	 *  
	 *  Schema: t{ g{ float }, l{ int } }
	 *  Query : t{ g{ var X } }
	 *  
	 *  Now we merely need to walk up the path [var X]->g->t in the query and in reverse down the schema t->g->[float]
	 *  var X is of type float.
	 *  
	 *  
	 *  It's desirable to do this differently IF DuraC supports complex composite types natively now.
	 *  Due to the uncertainties and changes in DuraC's grammar and feature set this code base grew with the
	 *  initial expectation it did not.
	 *  
	 *  Since we still don't have semantic analysis it doesn't make sense to fix this workaround with another workaround, instead
	 *  semantic analysis should yield structures that tell the types of variables etc. to be accessed in the desugaring transformations
	 *  and subsequent compilation...
	 *  
	 *   
	 */
	@SuppressWarnings(value = "unchecked")
	private void resolveExternalTypeReferences(CommonTree tree, HashMap<String, CommonTree> nameToTypeDefMap) {
		for (CommonTree schema : collectSchemata(tree)) {
			CommonTree schemaDef = (CommonTree) schema.getChild(0);
			
			for (CommonTree typeRef : collectNonBasicTypeReferences(schemaDef, nameToTypeDefMap)) {
				CommonTree resolved = nameToTypeDefMap.get(typeRef.getText());
				
				int originalIndex = typeRef.getChildIndex();
				int index = originalIndex;
				
				CommonTree typeRefParent = (CommonTree) typeRef.getParent();
				
				List<CommonTree> typeRefSiblings = typeRefParent.getChildren();
				
				for (Object c : resolved.getChildren()) {
					CommonTree child = (CommonTree) c;
					if (child.getType() == DuraParser.SUPPLEMENT) {
						break;
					}
					
					CommonTree childCopy = copyTree(child);
					
					typeRefSiblings.add(index, childCopy);
					index++;
				}
				
				typeRefParent.deleteChild(index);
				// The following line is not actually necessary since deleteChild() already calls this method.
				// http://www.antlr.org/api/Java/_base_tree_8java-source.html#l00154
				typeRefParent.freshenParentAndChildIndexes(originalIndex);

				addConstraints(schema, typeRefParent, resolved);
			}
		}
	}
	
	@SuppressWarnings(value="unchecked")
	private void addConstraints(CommonTree schemaOrTypeDef, CommonTree insertedInTerm, CommonTree addTypeDef) {
		CommonTree addSupplement = (CommonTree) addTypeDef.getChild(addTypeDef.getChildCount() - 1);
		if (addSupplement.getChildCount() == 0) {
			return;
		}
		
		CommonTree addTupleConstraints = (CommonTree) addSupplement.getChild(0);
		if (addTupleConstraints.getType() == DuraParser.TUPLE_CONSTRAINTS) {
			CommonTree constraintBasePath = new CommonTree(new CommonToken(DuraParser.PATH, null));
			
			CommonTree temp = insertedInTerm;
			ArrayList<CommonTree> tempList = new ArrayList<CommonTree>();
			while (temp.getType() == DuraParser.TERM) {
				tempList.add(0, temp);
				temp = (CommonTree) temp.getParent();
			}
			for (CommonTree term : tempList) {
				constraintBasePath.addChild(new CommonTree(new CommonToken(DuraParser.TERM, term.getText())));
			}
			
			CommonTree supplement = (CommonTree) schemaOrTypeDef.getChild(schemaOrTypeDef.getChildCount() - 1);
			CommonTree tupleConstraints = null;
			if (supplement.getChildCount() > 0 && supplement.getChild(0).getType() == DuraParser.TUPLE_CONSTRAINTS) {
				tupleConstraints = (CommonTree) supplement.getChild(0);
			} else {
				tupleConstraints = new CommonTree(new CommonToken(DuraParser.TUPLE_CONSTRAINTS, "TUPLE_CONSTRAINTS"));
				if (supplement.getChildCount() == 0) {
					supplement.addChild(tupleConstraints);
				} else {
					supplement.getChildren().add(0, tupleConstraints);
					supplement.freshenParentAndChildIndexes();
				}
			} 
			
			for (Object c : addTupleConstraints.getChildren()) {
				CommonTree constraint = (CommonTree) c;
				CommonTree newConstraint = copyTree(constraint);
				
				for (CommonTree path : collectPaths(newConstraint)) {
					CommonTree newPath = copyTree(constraintBasePath);
					
					path.getChildren().addAll(0, newPath.getChildren());
					path.freshenParentAndChildIndexes();
				}
				
				// TODO: This is slow and not necessary. It just minimizes the code output.
				boolean exists = false;
				if (tupleConstraints.getChildCount() > 0) {
					for (Object sc : tupleConstraints.getChildren()) {
						CommonTree tupleConstraintsChild = (CommonTree) sc;
					
						if (treesEqual(tupleConstraintsChild, newConstraint)) {
							exists = true;
						}
					}
				}
				if (!exists) {
					tupleConstraints.addChild(newConstraint);
				}
			}
		}
	}
	
	private LinkedList<CommonTree> collectPaths(CommonTree tree) {
		LinkedList<CommonTree> nodes = new LinkedList<CommonTree>();
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(tree);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.PATH) {
				nodes.add(node);
				continue;
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
		return nodes;
	}
	
	@SuppressWarnings(value = "unchecked")
	private void resolveInternalTypeReferences(HashMap<String, CommonTree> nameToTypeDefMap) {
		Collection<CommonTree> typeDefs = nameToTypeDefMap.values();
		
		// Time complexity estimates rely on the java implementation for LinkedList:
		// http://developer.classpath.org/doc/java/util/LinkedList-source.html
		final IdentityHashMap<CommonTree, LinkedList<CommonTree>> typeDefToRefsMap =
			new IdentityHashMap<CommonTree, LinkedList<CommonTree>>();
		for (CommonTree typeDef : typeDefs) {
			typeDefToRefsMap.put(typeDef, collectNonBasicTypeReferences(typeDef, nameToTypeDefMap));
		}
		
		// To speed things up the queue shall only contain custom type definitions that
		// contain custom type references, i.e. names referring to custom types.
		Queue<CommonTree> typeDefQueue = new LinkedList<CommonTree>();
		for (CommonTree typeDef : typeDefs) {
			if (typeDefToRefsMap.get(typeDef).size() > 0) {
				typeDefQueue.add(typeDef);
			}
		}
		
		int unexpandedTypeDefCount = typeDefQueue.size();
		int checkedTypeDefCount = 0;
		
		while(!typeDefQueue.isEmpty()) {
			CommonTree typeDef = (CommonTree) typeDefQueue.remove();
			
			boolean requeueTypeDef = false;
			
			// We use an explicit list iterator for constant time removal.
			ListIterator<CommonTree> typeRefsIter = typeDefToRefsMap.get(typeDef).listIterator(0);
			while (typeRefsIter.hasNext()) {
				CommonTree typeRef = typeRefsIter.next();
				CommonTree resolved = nameToTypeDefMap.get(typeRef.getText());
				
				// The size attribute is incremented/decremented during linked list construction/destruction.
				// The method isEmpty() relies on size, which is retrievable in constant time.
				if (typeDefToRefsMap.get(resolved).isEmpty()) {
					int originalIndex = typeRef.getChildIndex();
					int index = originalIndex;
					
					CommonTree typeRefParent = (CommonTree) typeRef.getParent();
					
					List<CommonTree> typeRefSiblings = typeRefParent.getChildren();
					
					for (Object c : resolved.getChildren()) {
						CommonTree child = (CommonTree) c;
						if (child.getType() == DuraParser.SUPPLEMENT) {
							break;
						}
						
						CommonTree childCopy = copyTree(child);
						
						typeRefSiblings.add(index, childCopy);
						index++;
					}
					
					typeRefParent.deleteChild(index);
					// The following line is not actually necessary since deleteChild() already calls this method.
					// http://www.antlr.org/api/Java/_base_tree_8java-source.html#l00154
					typeRefParent.freshenParentAndChildIndexes(originalIndex);
					
					addConstraints(typeDef, typeRefParent, resolved);
					
					// Remove this term paste from the linked list. This requires constant time.
					typeRefsIter.remove();
				} else {
					requeueTypeDef = true;
				}
			}
			
			if (requeueTypeDef) {
				typeDefQueue.add(typeDef);
			}
			
			checkedTypeDefCount++;
			if (checkedTypeDefCount > unexpandedTypeDefCount) {
				if (typeDefQueue.size() == unexpandedTypeDefCount) {
					StringBuilder sb = new StringBuilder();
					while (true) {
						CommonTree temp = (CommonTree) typeDefQueue.remove();
						sb.append("\"");
						sb.append(temp.getText());
						sb.append("\"");
						
						if (typeDefQueue.isEmpty()) {
							break;
						} else {
							sb.append(", ");
						}
					}
					
					throw new RuntimeException("Cycle in usage of type name(s) " + sb.toString() + ".");
				} else {
					checkedTypeDefCount = 0;
					unexpandedTypeDefCount = typeDefQueue.size();
				}
			}
		}
	}
	
	private LinkedList<CommonTree> collectNonBasicTypeReferences(CommonTree tree, HashMap<String, CommonTree> nameToTypeDefMap) {
		LinkedList<CommonTree> nodes = new LinkedList<CommonTree>();
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(tree);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.COMPOSITE_TYPE) {
				if (nameToTypeDefMap.get(node.getText()) == null) {
					throw new RuntimeException("No type definition for name " + node.getText() + " could be found.");
				}
				
				nodes.add(node);
				continue;
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
		return nodes;
	}
}
