package de.lmu.ifi.pms.cep.event_mill;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;

import org.simpleframework.xml.Serializer;
import org.simpleframework.xml.core.Persister;
import org.simpleframework.xml.strategy.CycleStrategy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.compilation.Compiler;
import de.lmu.ifi.pms.cep.compilation.CompilerFactory;
import de.lmu.ifi.pms.cep.compilation.StackCompilerFactory;
import de.lmu.ifi.pms.cep.event_mill.configuration.Configuration;
import de.lmu.ifi.pms.cep.event_mill.execution.engine.Engine;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgram;
import de.lmu.ifi.pms.cep.event_mill.execution.program.StreamBufferStore;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;

import de.lmu.ifi.pms.java.LoggingObject;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;



public class EventMill extends LoggingObject{
	
	Configuration configuration;
	StreamBufferStore<?, DTSA_Program> bufferStore;
	CompilerFactory<String[], TSA_Program> defaultToTSACompilerFactory;
	CompilerFactory<TSA_Program, DTSA_Program> defaultTsaDtsaCompilerFactory;
	
	public EventMill(){
		super();
	}
	
	public EventMill(final String configurationFilePath) throws IOException{
		this(Configuration.readFile(configurationFilePath));
	}
	
	public EventMill(final Configuration configuration) {
		this();
		this.config(configuration);
	}
	
	@SuppressWarnings("unchecked")
	void initBufferStore(){
		if(this.isConfigured()){
			this.bufferStore = (StreamBufferStore<?, DTSA_Program>)this.configuration.streamBufferStore.make();
			if(!this.bufferStore.isInizialized()){
				this.info("Initializing buffer store.");
				this.bufferStore.initialize();
				this.info("Initialized buffer store.");
			}
			else{
				this.debug("Buffer store is already initialized.");
			}
		}
		else{
			throw new IllegalStateException("Event-Mill is not yet configured.");
		}
	}
	
	void ensureBufferStore(){
		if(this.bufferStore == null){
			this.initBufferStore();
		}
	}
	
	public synchronized boolean isConfigured(){
		return this.configuration != null;
	}
	
	public synchronized Configuration configuration(){
		return this.configuration;
	}
	
	public synchronized void config(final String configFileName) throws IOException{
		this.config(Configuration.readFile(configFileName));
	}
	
	public synchronized void config(final Configuration configuration){
		this.configuration = configuration;
		if(this.bufferStore != null){
			this.bufferStore.close();
			this.bufferStore = null;
			this.initBufferStore();
		}
		this.defaultToTSACompilerFactory = configuration.defaultTSACompilerFactory;
		this.defaultTsaDtsaCompilerFactory = configuration.defaultDTSACompilerFactory;
		this.info("Set new configuration.");
	}
	
	public DTSA_Program compile(final String[] aliases, String namespace, final String simple_name, final String[] arguments) throws IOException{
		
		if(namespace == null){
			namespace = "";
		}
		
		Compiler<String[], DTSA_Program> toDTSACompiler;
		
		String[] compilerArguments = new String[arguments.length + 2];
		compilerArguments[0] = namespace;
		compilerArguments[1] = simple_name;
		System.arraycopy(arguments, 0, compilerArguments, 2, arguments.length);
		
		synchronized (this) {
			if(this.isConfigured()){
				ArrayList<CompilerFactory<?, ?>> compilers = new ArrayList<CompilerFactory<?, ?>>();
				for(int alias = 0; alias < aliases.length; ++alias){
					CompilerFactory<?,?> compiler = this.configuration.aliases.get(aliases[alias]);
					if(compiler == null){
						throw new IllegalArgumentException("Unknown compiler alias \"" + aliases[alias] + "\"");
					}
					else{
						compilers.add(compiler);
					}
				}
				
				if(compilers.size() == 0){
					compilers.add(this.defaultToTSACompilerFactory);
					assert this.defaultTsaDtsaCompilerFactory.sourceProgramType.isAssignableFrom(this.defaultToTSACompilerFactory.destinationProgramType);
				}
				
				assert compilers.size() > 0;
				
				if(DTSA_Program.class.isAssignableFrom(compilers.last().destinationProgramType)){
					//Nothing to do
				}
				else if(this.defaultTsaDtsaCompilerFactory.sourceProgramType.isAssignableFrom(compilers.last().destinationProgramType)){
					compilers.add(this.defaultTsaDtsaCompilerFactory);
				}
				else{
					if(this.defaultToTSACompilerFactory.sourceProgramType.isAssignableFrom(compilers.last().destinationProgramType)){
						compilers.add(this.defaultToTSACompilerFactory);
						compilers.add(this.defaultTsaDtsaCompilerFactory);
					}
					else{
						throw new IllegalArgumentException("Specified compiler stack does not define a valid ");
					}
				}

				CompilerFactory<String[], DTSA_Program> toDTSACompilerFactory = new StackCompilerFactory<String[], DTSA_Program>(String[].class, DTSA_Program.class, compilers.toArray(new CompilerFactory<?, ?>[compilers.size()]));
				toDTSACompiler = toDTSACompilerFactory.make();
			}
			else{
				throw new IllegalStateException("Event-Mill is not yet configured.");
			}
		}
		
		return toDTSACompiler.compile(compilerArguments);
	}
	
	public DTSA_Program compile(final String namespace, final String simple_name, final String[] arguments) throws IOException{ 
		return this.compile(new String[0], namespace, simple_name, arguments);
	}
	
	public DTSA_Program compile(final String full_name, final String... arguments) throws IOException{
		String[] splited_full_name = TSA_SyntaxPatterns.programFullName(full_name);
		return this.compile(splited_full_name[0], splited_full_name[1], arguments);
	}
	
	public void compile(final String destinationFileName, final String namespace, final String simple_name, final String[] arguments) throws IOException{
		DTSA_Program program_spec = this.compile(namespace, simple_name, arguments);
		this.save(program_spec, destinationFileName);
	}
	
	public DTSA_Program compile(final String namespace, final String simple_name, final File sourceFile) throws IOException{
		return this.compile(namespace, simple_name, new String[]{sourceFile.getPath()});
	}
	
	public void save(final DTSA_Program program_spec, final String destinationFileName) throws IOException{
		//works even if not configured
		Serializer serializer = new Persister(new CycleStrategy(), COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		File destinationFile = new File(destinationFileName).getAbsoluteFile();
		if(!destinationFile.exists()){
			File destinationDir = destinationFile.getParentFile();
			if(!destinationDir.exists()){
				if(destinationDir.mkdirs()){
					if(!destinationFile.createNewFile()){
						throw new IOException("Could not create file: " + destinationFile.getAbsolutePath());			
					}
				}
				else{
					throw new IOException("Could not create directory: " + destinationDir.getAbsolutePath());
				}
			}
		}
		Writer fileWriter = new BufferedWriter(new FileWriter(destinationFile));
		try {
			serializer.write(program_spec, fileWriter);
		} catch (Exception e) {
			throw new Error(e);
		}
		fileWriter.flush();
		fileWriter.close();
	}
	
	public DTSA_Program load(final String filename) throws IOException{
		//works even if not configured
		Persister persister = new Persister(new CycleStrategy(), COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		try {
			return persister.read(DTSA_Program.class, new File(filename));
		} 
		catch (RuntimeException e) {
			this.error("",e);
			throw e;
		} 
		catch (IOException e) {
			throw e;
		} 
		catch (Exception e) {
			this.error("",e);
			throw new RuntimeException(e);
		} 
		catch (Error e) {
			this.error("",e);
			throw e;
		}
	}
	
	public synchronized void addSpec(final String filename) throws IOException{
		DTSA_Program program_spec = this.load(filename);
		this.addSpec(program_spec);
	}
	
	public synchronized void addSpec(final DTSA_Program program_spec){
		this.ensureBufferStore();
		assert this.bufferStore != null;
		try{
			this.bufferStore.add(program_spec);
			this.info("Loaded program specification.");
		}
		catch (IllegalArgumentException e){							
			this.info("",e);
			throw e;
		}
		catch (RuntimeException e) {
			this.error("",e);
			throw e;
		}
		catch (Error e) {
			this.error("",e);
			throw e;
		}
	}

	public synchronized FixedArrayNavigableSet<String> availableProgramSpecifications(){
		this.ensureBufferStore();
		assert this.bufferStore != null;
		return this.bufferStore.availableProgramSpecifications();
	}
	
	public synchronized boolean hasSpec(final String full_name){
		this.ensureBufferStore();
		assert this.bufferStore != null;
		return this.bufferStore.existsProgSpec(full_name);
	}
	
	public synchronized void clearSpecs(){
		this.ensureBufferStore();
		assert this.bufferStore != null;
		this.info("Clearing program specifications.");
		
		this.bufferStore.clearProgramSpecs();
		
		this.info("Cleared program specifications.");
	}
	
	public synchronized void removeSpec(final String full_name){
		this.ensureBufferStore();
		assert this.bufferStore != null;
		this.info("Removing program specification.");

		this.bufferStore.removeProgramSpec(full_name);
			
		this.info("Removed program specification.");
	}
	
	public synchronized void instanciate(final String full_name, final String instance, final String[] parameters){
		String[] splited_name = TSA_SyntaxPatterns.programFullName(full_name);
		this.instanciate(splited_name[0], splited_name[1], instance, parameters);
	}
	
	public synchronized void instanciate(final String namespace, final String simple_name, final String instance, final String... parameters){
		this.ensureBufferStore();
		assert this.bufferStore != null;
		try{
			this.bufferStore.instanciate(namespace, simple_name, instance, parameters);
			this.info("Initialized program specification.");
		}
		catch (IllegalArgumentException e){							
			this.info("",e);
			throw e;
		}
		catch (RuntimeException e) {
			this.error("",e);
			throw e;
		}
		catch (Error e) {
			this.error("",e);
			throw e;
		}
	}
	
	public synchronized FixedArrayNavigableSet<String> availablePrograms(){
		this.ensureBufferStore();
		assert this.bufferStore != null;
		return this.bufferStore.availablePrograms();
	}
	
	public synchronized void clearProgs(){
		this.ensureBufferStore();
		assert this.bufferStore != null;
		this.info("Clearing program instances.");
		
		this.bufferStore.clearProgramInsts();
		
		this.info("Cleared program instances.");
	}
	
	public synchronized void removeProg(final String program_full_name, final String instance){
		this.removeProg(TSA_SyntaxPatterns.programInstFullName(program_full_name, instance));
	}
	
	public synchronized void removeProg(final String full_name){
		this.ensureBufferStore();
		assert this.bufferStore != null;
		this.info("Removing program instance.");

		this.bufferStore.removeProgramInst(full_name);
		
		this.info("Removed program instance.");
	}
	
	public synchronized ExecutableProgram<?,?,?> getProg(final String program_full_name, final String instance){		
		return this.getProg(TSA_SyntaxPatterns.programInstFullName(program_full_name, instance));
	}
	
	public synchronized ExecutableProgram<?,?,?> getProg(final String full_name){
		this.ensureBufferStore();
		assert this.bufferStore != null;
		return this.bufferStore.getProgram(full_name);
	}
	
	public Engine createEngine(final ExecutableProgram<?,?,?> program){
		return new Engine(program);
	}
	
	public synchronized void clear(){
		this.ensureBufferStore();
		assert this.bufferStore != null;
		this.info("Clearing program instances and program specifications.");
		
		this.bufferStore.clearProgramInsts();
		this.bufferStore.clearProgramSpecs();

		this.info("Cleared program instances and program specifications.");
	}
	
	public synchronized void uninstall(){
		this.ensureBufferStore();
		assert this.bufferStore != null;
		this.info("Uninstalling.");
		
		this.bufferStore.remove();
		this.bufferStore.close();
		this.bufferStore = null;

		this.info("Uninstalled.");
	}
	
	public synchronized void close(){
		if(this.bufferStore != null){
			this.bufferStore.close();
		}
	}
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(EventMill.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}
