package de.lmu.ifi.pms.cep.event_mill.console;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;

import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.Map;
import de.lmu.ifi.pms.java.util.map.compatibility.HashMap;
import de.lmu.ifi.pms.java.util.tupel.Single;


public abstract class AbstractConsoleThread extends Thread {
	
	static final Pattern variableDefinitionPattern = Pattern.compile("^[\\p{Blank}]*+set[\\p{Blank}]++%[{]((?:[^}]|\\\\[}])++)[}][\\p{Blank}]*+=[\\p{Blank}]*+(?:'((?:[^']|\\\\')*+)')?$");
	static final Pattern variableReplacementPattern = Pattern.compile("(?<!\\\\)%[{]((?:[^}]|\\\\[}])++)[}]");
	
	static final String pauseCommandRegEx = "pause(?:[ \\t]+(\\d++))?";
	static final Pattern pauseCommandPattern = Pattern.compile(pauseCommandRegEx);
	
	static final String injectCommandRegEx = "inject[ \\t]+(" + MainThread.fileNameRegEx + ")";
	static final Pattern injectCommandPattern = Pattern.compile(injectCommandRegEx);
	
	protected final ArrayList<Iterator<String>> bufferedCommands;
	protected boolean skipedCommands;
	
	protected final HashMap<String, String> variables;
	
	protected final BufferedReader inputStream;
	protected final PrintStream outputStream;
	
	public AbstractConsoleThread(final String name, final Iterator<String> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream) {		
		this(name, new HashMap<String, String>(), immediateCommands, inputStream, outputStream);
	}
	
	public AbstractConsoleThread(final String name, final Map<String, String, ?> variables, final Iterator<String> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream) {
		this(name, variables, new Single<Iterator<String>, Iterator<String>>(immediateCommands), inputStream, outputStream);
	}
	
	public AbstractConsoleThread(final String name, final Map<String, String, ?> variables, final Collection<Iterator<String>> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream) {
		super(name);
		
		this.variables = new HashMap<String, String>(variables);
		
		this.bufferedCommands = new ArrayList<Iterator<String>>(immediateCommands);
		this.inputStream = inputStream;
		this.outputStream = outputStream;
		
		this.skipedCommands = false;
	}

	@Override
	public void run() {

	}
	
	protected void inject(final String file) throws FileNotFoundException{
		this.inject(this.lineIterator(file));
	}
	
	protected void inject(final Iterator<String> injectedCommands){
		this.bufferedCommands.addLast(injectedCommands);
	}
	
	protected boolean isSetVariableCommand(final String command){
		return command.startsWith("set");
	}
	
	protected void resolveVariablesInSetVariable(final StringBuffer sb, final String setCommand){
		Matcher matcher = variableDefinitionPattern.matcher(setCommand);
		if(matcher.matches() && matcher.group(2) != null){
			sb.append(setCommand.substring(0, matcher.start(2)));
			this.resolveVariables(sb, matcher.group(2));
			sb.append(setCommand.substring(matcher.end(2)));
		}
		else{
			sb.append(setCommand);
		}
	}
	
	protected void resolveVariables(final StringBuffer sb, final String s){
		Matcher matcher = variableReplacementPattern.matcher(s);
		String var;
		String value;
		while(matcher.find()){
			var = matcher.group(1);
			value = this.getVariable(var);
			if(value != null){
				matcher.appendReplacement(sb, Matcher.quoteReplacement(value));
			}
			else{
				throw new RuntimeException("Variable %{" + var + "} is not defined.");//FIXME better exception
			}
		}
		matcher.appendTail(sb);
	}
	
	protected String resolvedCommand(final String command){
		StringBuffer sb = new StringBuffer();
		if(this.isSetVariableCommand(command)){
			this.resolveVariablesInSetVariable(sb, command);
		}
		else{
			this.resolveVariables(sb, command);
		}
		return sb.toString();
	}
	
	protected String nextCommand(){
		String command;
		String resolvedCommand;
		while(!this.bufferedCommands.isEmpty() && !this.bufferedCommands.last().hasNext()){
			this.bufferedCommands.removeLast();
		}
		if(!this.bufferedCommands.isEmpty()){
			assert this.bufferedCommands.last().hasNext();
			command = this.bufferedCommands.last().next();
			resolvedCommand = this.resolvedCommand(command);
			this.outputStream.println(resolvedCommand);
		}
		else{
			try {
				if(!this.inputStream.ready()){
					this.outputStream.print(">");
				}
				command = this.inputStream.readLine();
				if(command == null){
					this.info("End of input stream has been reached.");
					this.inputStream.close();
					throw new EndOfInputException("End of input stream has been reached.");
				}
				else{
					resolvedCommand = this.resolvedCommand(command);
				}
			} catch (IOException e) {
				
				this.error("Exception when reading from input.", e);
				
				throw new RuntimeException(e);
				//TODO better exception
			}
		}
		assert command != null;
		return resolvedCommand;
	}
	
	protected boolean skipedCommands(){
		return this.skipedCommands;
	}
	
	protected void skipBufferedCommands(){
		this.bufferedCommands.clear();
		this.skipedCommands = true;		
	}
	
	protected void skipRemainigCommands(){
		this.skipBufferedCommands();
		try {
			while(this.inputStream.ready()){
				this.inputStream.readLine();
			}
		} catch (IOException e) {
			
			this.error("Exception when reading from input.", e);
			
			throw new RuntimeException(e);
			//TODO better exception
		}
	}
	
	protected void afterSetVariable(final String var, final String value){
		
	}
	
	protected void setVariable(final String var, final String value){
		if(value == null){
			this.variables.removeKey(var);
		}
		else{
			this.variables.set(var, value);
		}
		this.afterSetVariable(var, value);
	}
	
	protected boolean isSetVariable(final String var){
		return this.variables.has(var);
	}
	
	protected String getVariable(final String var) {
		return this.variables.get(var);
	}
	
	protected void pause(){
		try {
			this.inputStream.read();
		} catch (IOException e) {
			this.debug("Exception when reading from input.", e);
		}
	}
	
	protected void pause(final long milies){
		try {//TODO Loop?
			Thread.sleep(milies);
		} catch (InterruptedException e) {
			this.info("", e);
		}
	}
	
	public Iterator<String> lineIterator(final String file) throws FileNotFoundException{
		return this.lineIterator(new BufferedReader(new FileReader(file)));
	}
	
	public Iterator<String> lineIterator(final BufferedReader fileReader){
		return new Iterator<String>() {
			
			String next;
			
			@Override
			public boolean hasNext() {
				if(this.next == null){
					try {
						this.next = fileReader.readLine();
					} catch (IOException e) {
						throw new IllegalStateException("I/O error while reading next line.", e);
					}
				}
				if(this.next == null){
					try {
						fileReader.close();
					} catch (IOException e) {
						AbstractConsoleThread.this.warn("End of input stream has been reached.", e);
					}
					return false;
				}
				else{
					return true;
				}
			}
			
			@Override
			public String next() {
				if(this.hasNext()){
					String result = this.next;
					this.next = null;
					return result;
				}
				else{
					throw new NoSuchElementException();
				}
			}
			
			@Override
			public void remove() {
				throw new UnsupportedOperationException();
			}
			
		};
	}
	
	public static class EndOfInputException extends NoSuchElementException{

		public EndOfInputException() {
			super();
		}

		public EndOfInputException(final String s) {
			super(s);
		}
		
	}
	
	//logging	
	
	protected abstract Logger logger();

	protected void trace(final String msg) {
		this.logger().trace(msg);
	}

	protected void trace(final String format, final Object arg) {
		this.logger().trace(format, arg);
	}

	protected void trace(final String format, final Object arg1, final Object arg2) {
		this.logger().trace(format, arg1, arg2);
	}

	protected void trace(final String format, final Object... argArray) {
		this.logger().trace(format, argArray);
	}

	protected void trace(final String msg, final Throwable t) {
		this.logger().trace(msg, t);
	}

	protected void debug(final String msg) {
		this.logger().debug(msg);
	}

	protected void debug(final String format, final Object arg) {
		this.logger().debug(format, arg);
	}

	protected void debug(final String format, final Object arg1, final Object arg2) {
		this.logger().debug(format, arg1, arg2);
	}

	protected void debug(final String format, final Object... argArray) {
		this.logger().debug(format, argArray);
	}

	protected void debug(final String msg, final Throwable t) {
		this.logger().debug(msg, t);
	}

	protected void info(final String msg) {
		this.logger().info(msg);
	}

	protected void info(final String format, final Object arg) {
		this.logger().info(format, arg);
	}

	protected void info(final String format, final Object arg1, final Object arg2) {
		this.logger().info(format, arg1, arg2);
	}

	protected void info(final String format, final Object... argArray) {
		this.logger().info(format, argArray);
	}

	protected void info(final String msg, final Throwable t) {
		this.logger().info(msg, t);
	}

	protected void warn(final String msg) {
		this.logger().warn(msg);
	}

	protected void warn(final String format, final Object arg) {
		this.logger().warn(format, arg);
	}

	protected void warn(final String format, final Object arg1, final Object arg2) {
		this.logger().warn(format, arg1, arg2);
	}

	protected void warn(final String format, final Object... argArray) {
		this.logger().warn(format, argArray);
	}

	protected void warn(final String msg, final Throwable t) {
		this.logger().warn(msg, t);
	}

	protected void error(final String msg) {
		this.logger().error(msg);
	}

	protected void error(final String format, final Object arg) {
		this.logger().error(format, arg);
	}

	protected void error(final String format, final Object arg1, final Object arg2) {
		this.logger().error(format, arg1, arg2);
	}

	protected void error(final String format, final Object... argArray) {
		this.logger().error(format, argArray);
	}

	protected void error(final String msg, final Throwable t) {
		this.logger().error(msg, t);
	}

}
