package de.lmu.ifi.pms.cep.event_mill.execution.program;

import static de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgramState.*;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArraySet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;


public abstract class ExecutableProgram<StoreType extends StreamBufferStore<StoreType, ?>, BufferType extends StreamBuffer<BufferType,QueryType,?>, QueryType extends ExecutableQuery<?, BufferType>>
{
	
	public final String namespace;
	public final String simple_name;
	public final String instance;
	public final String fullName;
	
	protected ExecutableProgramState state;
	
	public final StoreType bufferStore;
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(ExecutableProgram.class);
	public final Marker marker;
	
	protected final Logger progLogger;

	public ExecutableProgram(final String namespace, final String simple_name, final String instance, final StoreType bufferStore)
	{
		super();
		assert namespace.matches(TSA_SyntaxPatterns.ProgramNamespaceRegEx);
		assert simple_name.matches(TSA_SyntaxPatterns.ProgramNameRegEx);
		assert instance.matches(TSA_SyntaxPatterns.program_instance_name_RegEx);
		
		this.namespace = namespace;
		this.simple_name = simple_name;
		this.instance = instance;
		this.fullName = TSA_SyntaxPatterns.programInstFullName(this.namespace, this.simple_name, this.instance);
		
		this.bufferStore = bufferStore;
		
		this.marker = MarkerFactory.getMarker(this.name());
		
		this.progLogger = LoggerFactory.getLogger(this.name());
		
		this.state = null;
	
	}
	
	public final String namespace(){
		return this.namespace;
	}
	
	public final String name(){
		return this.fullName;
	}
	
	public final String simpleName(){
		return this.simple_name;
	}

	public abstract FixedEnumMap<DTSA_BufferType, ? extends FixedArrayNavigableMap<String, BufferType>> StreamBuffers();
	
	public abstract FixedEnumMap<DTSA_QueryType, FixedArraySet<QueryType>> Queries();
	
	protected abstract ExecutableProgramState readState();
	
	protected void writeState(final ExecutableProgramState state){
		
	}
	
	protected abstract boolean aquireLock();
	
	protected abstract void releaseLock();

	//should only be called in the Engine methods
	public synchronized void lock(){		
		if(this.state == null){
			if(this.aquireLock()){
				this.state = this.readState();
				assert this.state != null;
				this.info("{}",this.state);
				this.notifyAll();
				this.info("Locked");
			}
			else{
				IllegalStateException e = new IllegalStateException("Program could not be locked.");
				this.debug("Program could not be locked. aquireLock() did not succeed.", e);
				throw e;
			}
		}
		else{
			IllegalStateException e = new IllegalStateException("Program is already locked.");
			this.debug("lock() called in when program was already locked.", e);
			throw e;
		}
	}
	
	public FixedArraySet<QueryType> topologicalOrder(final FixedArraySet<QueryType> queries){
		return defaultTopologicalOrder(queries, this.StreamBuffers().values());
	}
	
	public static <QueryType extends ExecutableQuery<?, ?>> FixedArraySet<QueryType> defaultTopologicalOrder(final FixedArraySet<QueryType> queries, final FixedArray<? extends FixedArrayNavigableMap<String, ? extends StreamBuffer<?, ?, ?>>> buffers){
		@SuppressWarnings("unchecked")
		QueryType[] queriesArray = queries.toArray((QueryType[]) new ExecutableQuery<?,?>[queries.size()]);
		ArrayList<QueryType> orderedQueries = new ArrayList<QueryType>();
		boolean[] isPositioned = new boolean[queriesArray.length];
		boolean[] isOnStack = new boolean[queriesArray.length];
		
		
		int numberOfBufferTypes = buffers.size();		
		@SuppressWarnings("unchecked")
		FixedArrayNavigableMap<String,ArrayList<Integer>>[] queriesByOutputArray = (FixedArrayNavigableMap<String,ArrayList<Integer>>[]) new FixedArrayNavigableMap<?,?>[numberOfBufferTypes];
		for(int buffer_type = 0; buffer_type < numberOfBufferTypes; ++buffer_type){
			FixedArrayNavigableSet<String> bufferNames = buffers.get(buffer_type).keys();
			@SuppressWarnings("unchecked")
			ArrayList<Integer>[] queryLists = (ArrayList<Integer>[]) new ArrayList<?>[bufferNames.size()];
			for(int buffer = 0; buffer < queryLists.length; ++buffer){
				queryLists[buffer] = new ArrayList<Integer>();
			}
			queriesByOutputArray[buffer_type] = new FixedArrayNavigableMap<String, ArrayList<Integer>>(bufferNames, new FixedArray<ArrayList<Integer>>(false, queryLists));
		}
		
		FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, ArrayList<Integer>>> queriesByOutput = new FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String,ArrayList<Integer>>>(DTSA_BufferType.class, queriesByOutputArray);
		
		for(int queryIndex = 0; queryIndex < queriesArray.length; ++queryIndex){
			for(StreamBuffer<?, ?, ?> output: queriesArray[queryIndex].outputs().values()){
				queriesByOutput.get(output.type).get(output.full_name).add(queryIndex);
			}
		}
		
		for(int queryIndex = 0; queryIndex < queriesArray.length; ++queryIndex){
			defaultTopologicalOrder(queryIndex, isPositioned, isOnStack, queriesByOutput, queriesArray, orderedQueries);
		}
		
		return new FixedArraySet<QueryType>(false, orderedQueries.toArray(queriesArray));
	}
	
	protected static <QueryType extends ExecutableQuery<?, ?>> void defaultTopologicalOrder(final int queryIndex, final boolean[] isPositioned, final boolean[] isOnStack, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, ArrayList<Integer>>> queriesByOutput, final QueryType[] queriesArray, final ArrayList<QueryType> orderedQueries){
		if(!isPositioned[queryIndex]){
			if(isOnStack[queryIndex]){
				throw new RuntimeException("Detected cyclic dependency on static queries.");
			}
			else{
				isOnStack[queryIndex] = true;
				QueryType query = queriesArray[queryIndex];
				for(StreamBuffer<?, ?, ?> input: query.inputs().values()){
					for(Integer predecessorQueryIndex: queriesByOutput.get(input.type).get(input.full_name)){
						defaultTopologicalOrder(predecessorQueryIndex, isPositioned, isOnStack, queriesByOutput, queriesArray, orderedQueries);
					}
				}
				
				orderedQueries.add(query);
				isPositioned[queryIndex] = true;
				isOnStack[queryIndex] = false;
			}
		}
	}
	
	public synchronized void init() {
		this.checkLock();
		if(this.state == CREATED){
			//TODO do at another place/way? -> assynchrone / distribution
			
			for(ExecutableQuery<?, ?> staticQuery: this.topologicalOrder(this.Queries().get(DTSA_QueryType.STATIC))){
				staticQuery.execute();
			}
			for(ExecutableQuery<?, ?> staticGarbageQuery: this.Queries().get(DTSA_QueryType.GARBAGE_STATIC)){
				staticGarbageQuery.execute();
			}
			this.state = INITIALIZED;
			this.writeState(this.state);
			this.notifyAll();
			this.info("{}",this.state);
		}
		else{
			IllegalStateException e = new IllegalStateException("init() called in state " + this.state);
			this.debug("init() called in wrong state! Raising exeption.", e);
			throw e;
			//TODO better exception
		}
	}
	
	public synchronized boolean isInitialized(){
		return this.state != CREATED;
	}
	
	public synchronized boolean isLocked(){
		return this.state != null;
	}
	
	public synchronized void checkLock(){
		if(this.state == null){
			throw new IllegalStateException("The program is not locked.");
		}
	}
	
	//should only be called in the Engine methods
	public synchronized void unlock(){
		this.checkLock();
		assert this.state != null;
		if(this.state == CREATED | this.state == INITIALIZED | this.state == STOPPED){
			this.releaseLock();
			this.state = null;
			this.notifyAll();
			this.info("Unlocked");
		}
		else{
			IllegalStateException e = new IllegalStateException("unlock() called in state " + this.state.name());
			this.debug("unlock() called in wrong state! Raising exeption.", e);
			throw e;
		}
	}
	
	protected synchronized void atStartInit(){
		
	}
	
	public synchronized void run(){
		this.checkLock();
		if(this.state == INITIALIZED | this.state == STOPPED){
			if(this.state == INITIALIZED){
				this.atStartInit();
				
				for(ExecutableQuery<?, ?> initQuery: this.Queries().get(DTSA_QueryType.INIT)){
					initQuery.execute();
				}
				for(ExecutableQuery<?, ?> initGarbageQuery: this.Queries().get(DTSA_QueryType.GARBAGE_INIT)){
					initGarbageQuery.execute();
				}
				this.state = STOPPED;
				this.writeState(this.state);
			}
						
			this.state = RUNNING;
			this.writeState(this.state);
			this.notifyAll();
			this.info("{}",this.state);
		}
		else{
			IllegalStateException e = new IllegalStateException();
			this.debug("run() called in wrong state! Raising exeption.", e);
			throw e;
		}
	}
	
	public synchronized void stop(){
		this.checkLock();
		if(this.state == RUNNING){
			this.state = STOPPED;
			this.writeState(this.state);
			this.notifyAll();
			this.info("{}",this.state);
		}
		else{
			IllegalStateException e = new IllegalStateException();
			this.debug("stop() called in wrong state! Raising exeption.", e);
			throw e;
		}
	}
	
	public synchronized ExecutableProgramState State(){
		this.checkLock();
		return this.state;
	}
	
	
	//logging
	
	protected Logger logger(){
		return logger;
	}

	protected final void trace(final String msg) {
		this.logger().trace(this.marker, msg);
	}

	protected final void trace(final String format, final Object arg) {
		this.logger().trace(this.marker, format, arg);
	}

	protected final void trace(final String format, final Object arg1, final Object arg2) {
		this.logger().trace(this.marker, format, arg1, arg2);
	}

	protected final void trace(final String format, final Object[] argArray) {
		this.logger().trace(this.marker, format, argArray);
	}

	protected final void trace(final String msg, final Throwable t) {
		this.logger().trace(this.marker, msg, t);
	}

	protected final void debug(final String msg) {
		this.logger().debug(this.marker, msg);
	}

	protected final void debug(final String format, final Object arg) {
		this.logger().debug(this.marker, format, arg);
	}

	protected final void debug(final String format, final Object arg1, final Object arg2) {
		this.logger().debug(this.marker, format, arg1, arg2);
	}

	protected final void debug(final String format, final Object[] argArray) {
		this.logger().debug(this.marker, format, argArray);
	}

	protected final void debug(final String msg, final Throwable t) {
		this.logger().debug(this.marker, msg, t);
	}

	protected final void info(final String msg) {
		this.logger().info(this.marker, msg);
	}

	protected final void info(final String format, final Object arg) {
		this.logger().info(this.marker, format, arg);
	}

	protected final void info(final String format, final Object arg1, final Object arg2) {
		this.logger().info(this.marker, format, arg1, arg2);
	}

	protected final void info(final String format, final Object[] argArray) {
		this.logger().info(this.marker, format, argArray);
	}

	protected final void info(final String msg, final Throwable t) {
		this.logger().info(this.marker, msg, t);
	}

	protected final void warn(final String msg) {
		this.logger().warn(this.marker, msg);
	}

	protected final void warn(final String format, final Object arg) {
		this.logger().warn(this.marker, format, arg);
	}

	protected final void warn(final String format, final Object arg1, final Object arg2) {
		this.logger().warn(this.marker, format, arg1, arg2);
	}

	protected final void warn(final String format, final Object[] argArray) {
		this.logger().warn(this.marker, format, argArray);
	}

	protected final void warn(final String msg, final Throwable t) {
		this.logger().warn(this.marker, msg, t);
	}

	protected final void error(final String msg) {
		this.logger().error(this.marker, msg);
	}

	protected final void error(final String format, final Object arg) {
		this.logger().error(this.marker, format, arg);
	}

	protected final void error(final String format, final Object arg1, final Object arg2) {
		this.logger().error(this.marker, format, arg1, arg2);
	}

	protected final void error(final String format, final Object[] argArray) {
		this.logger().error(this.marker, format, argArray);
	}

	protected final void error(final String msg, final Throwable t) {
		this.logger().error(this.marker, msg, t);
	}

	
}
