package de.lmu.ifi.pms.cep.event_mill.sql.io;

import java.lang.reflect.Array;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.io.stream.SinkStream;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_EXCEPTIONS;
import de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.collection.array.Arrays;


public class SQL_SinkStream extends SinkStream{
	
	final SQL_EngineIO io;
	final String buffer_name;
	final String[] writeSql;
	final String readBufferContentSql;
	final String readBufferContentMessage;
	final PreparedStatement[] write;
	int tupleIndex;
	int capacity;
	final Object[] valueBuffer;
	final BasicTypeID[] typeArray;
	
	protected Statement loggingStatement;
	
	public SQL_SinkStream(final String buffer_name, final SQL_EngineIO io) {
		super(buffer_name, io.attributeTypes(DTSA_BufferType.INPUT, buffer_name));
		this.io = io;
		this.buffer_name = buffer_name;
		this.writeSql = new String[31];
		
		if(this.logger().isTraceEnabled()){
			this.readBufferContentSql = this.io.readCurrentBufferContentSQL(DTSA_BufferType.INPUT, buffer_name);
			this.readBufferContentMessage = "Snapshot of table \"" + this.io.inputSchema() + "\".\"" + this.io.bufferNameSQLMapping(DTSA_BufferType.INPUT, buffer_name) + "\" after the execution of the statements:\n{}";
		}
		else{
			this.readBufferContentSql = null;
			this.readBufferContentMessage = null;
		}
		
		this.write = new PreparedStatement[31];
		
		this.tupleIndex = -1;
		this.capacity = 1;
		this.valueBuffer = new Object[this.attributes.size()];
		this.typeArray = new BasicTypeID[this.attributes.size()];
		int initialCapacity = 1;
		BasicTypeID attributeType;
		for(int attribute = 0; attribute < this.valueBuffer.length; ++attribute){
    		attributeType = this.types.values().get(attribute);
    		this.typeArray[attribute] = attributeType;
    		this.valueBuffer[attribute] = this.createArray(attributeType, initialCapacity);
		}
		
		this.initStatements();
	}
	
	Object createArray(final BasicTypeID type, final int length){
		switch (type) {
		case BOOLEAN:
			return Array.newInstance(Boolean.TYPE, length);
		case STRING:
			return Array.newInstance(String.class, length);
		case INT:
			return Array.newInstance(Integer.TYPE, length);
		case LONG:
			return Array.newInstance(Long.TYPE, length);
		case FLOAT:
			return Array.newInstance(Float.TYPE, length);
		case DOUBLE:
			return Array.newInstance(Double.TYPE, length);
		case TIMESTAMP:
			return Array.newInstance(Long.TYPE, length);
		case DURATION:
			return Array.newInstance(Long.TYPE, length);
		case IDENTIFIER:
			return Array.newInstance(Long.TYPE, length);
		case REFERENCE: throw new IllegalArgumentException("Type \"reference\" not supported.");
		default:
			throw new IllegalArgumentException("Type \"" + type.name() + "\" not supported.");
		}
	}
	
	void reconnect(){
		assert this.write[0] != null;
		if(this.write[0] != null){
			try{			
				this.io.reconnect(this.write[0].getConnection());
				for(int expIndex = 0; expIndex < 31; ++expIndex){
					try{ this.write[expIndex].close(); } catch (SQLException e) { this.debug("", e); }
					this.write[expIndex] = null;
				}
			} catch (SQLException e) {
				this.error("", e);
				throw new RuntimeException(e);
			}
		}
		if(this.loggingStatement != null){
			try{ this.loggingStatement.close(); } catch (SQLException e) { this.debug("", e); }
			this.loggingStatement = null;
		}
	}
	
	protected void initStatements(){
		while(true){
			try{
				if(this.logger().isDebugEnabled()){
					if(this.loggingStatement != null){
						try{this.loggingStatement.close();} catch (SQLException e){}
					}
					this.loggingStatement = this.io.createStatement();
				}
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect();
					continue;
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	protected void initStatement(final int expSize){
		assert 0 <= expSize & expSize < 31;
		while(true){
			try{
				if(this.writeSql[expSize] == null){
					this.writeSql[expSize] = this.io.inputSQL(this.buffer_name, 1 << expSize);
				}
				this.write[expSize] = this.io.prepareStatement(this.writeSql[expSize]);
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect();
					this.initStatements();
					continue;
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	@Override
	public boolean nextBlock(){
		if(this.isOpen){
			if(this.tupleIndex >= 0){
				BasicTypeID attributeType;
				int tuple = 0;
				int numberOfRemainingTuples = this.tupleIndex + 1;
				
				while(numberOfRemainingTuples > 0){
				
					int expIndex = 31 - Integer.numberOfLeadingZeros(numberOfRemainingTuples);
					expIndex = (expIndex > 8 ? 8 : expIndex);
					assert expIndex >= 0;
					int blockSize = 1 << expIndex;
					int endIndex = tuple + blockSize;
					//numberOfRemainingTuples ^= blockSize;
					numberOfRemainingTuples -= blockSize;
						
						while(true){
							try {
								PreparedStatement write = this.write[expIndex];
								if(write == null){
									this.initStatement(expIndex);
									write = this.write[expIndex];
								}
								int parameter = 0;
								
								for(; tuple < endIndex; ++tuple){
									assert tuple <= this.tupleIndex;
								
							    	for (int attribute = 1; attribute <= this.valueBuffer.length; ++attribute){
							    		
							    		++parameter;
							    		
							    		attributeType = this.typeArray[attribute - 1];
							    		switch (attributeType) {
										case BOOLEAN:
											write.setBoolean(parameter, Array.getBoolean(this.valueBuffer[attribute - 1], tuple));
											break;
										case STRING:
											write.setString(parameter, ((String[])this.valueBuffer[attribute - 1])[tuple]);
											break;
										case INT:
											write.setInt(parameter, Array.getInt(this.valueBuffer[attribute - 1], tuple));
											break;
										case LONG:
											write.setLong(parameter, Array.getLong(this.valueBuffer[attribute - 1], tuple));
											break;				
										case FLOAT:
											write.setFloat(parameter, Array.getFloat(this.valueBuffer[attribute - 1], tuple));
											break;
										case DOUBLE:
											write.setDouble(parameter, Array.getDouble(this.valueBuffer[attribute - 1], tuple));
											break;
										case TIMESTAMP:
											write.setLong(parameter, Array.getLong(this.valueBuffer[attribute - 1], tuple));
											break;				
										case DURATION:
											write.setLong(parameter, Array.getLong(this.valueBuffer[attribute - 1], tuple));
											break;				
										case IDENTIFIER:
											write.setLong(parameter, Array.getLong(this.valueBuffer[attribute - 1], tuple));
											break;
										case REFERENCE: throw new IllegalArgumentException("Type \"reference\" not supported.");
										default:
											break;
										}
							    	}
								}
								
								assert parameter == blockSize * this.valueBuffer.length;
								
								write.executeUpdate();
								try{ write.clearParameters(); } catch (SQLException e) { }
								break;
							} catch (SQLException e) {
								if(SQL_EXCEPTIONS.isConnectionFailure(e)){
									this.debug("Connection failure. Trying again.", e);
									this.reconnect();//FIXME current connection as parameter, otherwise cicular reconnect over several result streams
									this.initStatements();
									continue;
								}
								else if(SQL_EXCEPTIONS.isRollback(e)){
									this.debug("Statement rolled back. Trying again.", e);
								}
								else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
									this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
									this.initStatement(expIndex);
								}
								else{
									throw new RuntimeException(e);
								}
							}
						}
				}
				this.clear();
			}
			
			
			if(this.logger().isTraceEnabled()){
				
				while(true){
					try{
						assert this.loggingStatement != null;
						ResultSet tuples = this.loggingStatement.executeQuery(this.readBufferContentSql);
						String tuplesTable = SQL_QueryBuildHelper.toString(tuples, -1);
						this.trace(this.readBufferContentMessage, tuplesTable);
						break;
					}
					catch(SQLException e){
						if(SQL_EXCEPTIONS.isConnectionFailure(e)){
							this.debug("Connection failure. Trying again.", e);
							this.reconnect();
							this.initStatements();
							continue;
						}
						else if(SQL_EXCEPTIONS.isRollback(e)){
							this.debug("Debug statement rolled back. Trying again.", e);
						}
						else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
							this.debug("Handle of statement is invalid. Repreparing statement and trying again.", e);
							this.initStatements();
						}
						else{
							throw new RuntimeException(e);
						}
					}				
				}
				
			}
			
			
			return true;
		}
		else{
			return false;
		}
	}
	
	public void clear(){
		this.tupleIndex = -1;
		BasicTypeID attributeType;
		for (int attribute = 0; attribute < this.valueBuffer.length; ++attribute){
    		attributeType = this.typeArray[attribute];
    		switch (attributeType) {
			case BOOLEAN:
				Arrays.fill((boolean[]) this.valueBuffer[attribute], false);
				break;
			case STRING:
				Arrays.fill((String[]) this.valueBuffer[attribute], null);
				break;
			case INT:
				Arrays.fill((int[]) this.valueBuffer[attribute], 0);
				break;
			case LONG:
				Arrays.fill((long[]) this.valueBuffer[attribute], 0l);
				break;				
			case FLOAT:
				Arrays.fill((float[]) this.valueBuffer[attribute], 0f);
				break;
			case DOUBLE:
				Arrays.fill((double[]) this.valueBuffer[attribute], 0d);
				break;
			case TIMESTAMP:
				Arrays.fill((long[]) this.valueBuffer[attribute], 0l);
				break;				
			case DURATION:
				Arrays.fill((long[]) this.valueBuffer[attribute], 0l);
				break;				
			case IDENTIFIER:
				Arrays.fill((long[]) this.valueBuffer[attribute], 0l);
				break;
			case REFERENCE: throw new IllegalArgumentException("Type \"reference\" not supported.");
			default:
				break;
			}
    	}
	}
	
	@Override
	public boolean next() {
		++this.tupleIndex;
		if(this.tupleIndex >= this.capacity){
			int oldCapacity = this.capacity;
			int newCapacity = oldCapacity * 2;
			for(int attribute = 0; attribute < this.valueBuffer.length; ++attribute){
				Object oldArray = this.valueBuffer[attribute];
				Object newArray = Array.newInstance(oldArray.getClass().getComponentType(), newCapacity);
				System.arraycopy(oldArray, 0, newArray, 0, oldCapacity);
	    		this.valueBuffer[attribute] = newArray;
			}
			this.capacity = newCapacity;
		}
		return true;
	}
	
	@Override
	public void setString(final int columnIndex, final String value) {		
		Array.set(this.valueBuffer[columnIndex - 1], this.tupleIndex, value);
	}

	@Override
	public void setBoolean(final int columnIndex, final boolean value) {
		Array.setBoolean(this.valueBuffer[columnIndex - 1], this.tupleIndex, value);
	}

	@Override
	public void setByte(final int columnIndex, final byte value) {
		Array.setByte(this.valueBuffer[columnIndex - 1], this.tupleIndex, value);
	}

	@Override
	public void setShort(final int columnIndex, final short value) {
		Array.setShort(this.valueBuffer[columnIndex - 1], this.tupleIndex, value);
	}

	@Override
	public void setInt(final int columnIndex, final int value) {
		Array.setInt(this.valueBuffer[columnIndex - 1], this.tupleIndex, value);
	}

	@Override
	public void setLong(final int columnIndex, final long value) {
		Array.setLong(this.valueBuffer[columnIndex - 1], this.tupleIndex, value);
	}

	@Override
	public void setTimestamp(final int columnIndex, final long value) {
		Array.setLong(this.valueBuffer[columnIndex - 1], this.tupleIndex, value);
	}

	@Override
	public void setDuration(final int columnIndex, final long value) {
		Array.setLong(this.valueBuffer[columnIndex - 1], this.tupleIndex, value);
	}

	@Override
	public void setIdentifier(final int columnIndex, final long value) {
		Array.setLong(this.valueBuffer[columnIndex - 1], this.tupleIndex, value);
	}

	@Override
	public void setFloat(final int columnIndex, final float value) {
		Array.setFloat(this.valueBuffer[columnIndex - 1], this.tupleIndex, value);
	}

	@Override
	public void setDouble(final int columnIndex, final double value) {
		Array.setDouble(this.valueBuffer[columnIndex - 1], this.tupleIndex, value);
	}

	
	@Override
	public void close() {
		super.close();
		for(int expIndex = 0; expIndex < 31; ++expIndex){
			try{if(this.write[expIndex] != null){ this.write[expIndex].close(); } } catch (SQLException e){}
		}
	}

	
//	public boolean isFirst() {
//
//	}
//
//	
//	public boolean isLast() {
//
//	}

	
//	public void setFetchSize(final int rows) {
//
//	}
//
//	
//	public int getFetchSize() {
//
//	}
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(SQL_SinkStream.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}