package de.lmu.ifi.pms.cep.tsa;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_ComprehensiveException;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


@Root(name="program")
public class TSA_Program {
	
	public final FixedCollection<String> origin;
	
	@ElementList(name="origin", inline=true)
	java.util.ArrayList<String> getOriginList(){
		return new java.util.ArrayList<String>(this.origin.javaCollection());
	}
	
	@Attribute(name="namespace")
	public final String namespace;
	@Attribute(name="name")
	public final String simple_name;
	@Element(name="description", data=true)
	public final String description;
	
	public final FixedArrayNavigableMap<String, TSA_TemporalStream> temporalStreams;
	
	@ElementList(name="temporal-streams", entry="temporal-stream", type=TSA_TemporalStream.class)
	java.util.ArrayList<TSA_TemporalStream> getTemporalStreamList(){
		return new java.util.ArrayList<TSA_TemporalStream>(this.temporalStreams.values().javaCollection());
	}
	
	public final FixedArray<TSA_Query> queries;
	
	@ElementList(name="queries", entry="query", type=TSA_Query.class)
	java.util.ArrayList<TSA_Query> getQueryList(){
		return new java.util.ArrayList<TSA_Query>(this.queries.javaCollection());
	}
	
	public TSA_Program(final String namespace, final String simple_name, final String description, final FixedArray<TSA_TemporalStream> temporalStreams, final TSA_Query[] queries, final FixedCollection<String> origin)  throws TSA_Exception{
		this(namespace, simple_name, description, temporalStreams, new FixedArray<TSA_Query>(queries), origin);
	}
	
	public TSA_Program(final String namespace, final String simple_name, final String description, final java.util.ArrayList<TSA_TemporalStream> temporalStreams, final java.util.ArrayList<TSA_Query> queries, final FixedCollection<String> origin)  throws TSA_Exception{
		this(namespace, simple_name, description,  new FixedArray<TSA_TemporalStream>(temporalStreams), new FixedArray<TSA_Query>(queries), origin);
	}

	public TSA_Program(final String namespace, final String simple_name, final String description, final FixedArray<TSA_TemporalStream> temporalStreams, final FixedArray<TSA_Query> queries, final FixedCollection<String> origin)  throws TSA_Exception{
		super();
		ArrayList<TSA_Exception> exceptions = new ArrayList<TSA_Exception>();
		
		if(TSA_SyntaxPatterns.isValidNamespace(namespace)){
			this.namespace = namespace;
		}
		else{
			throw new TSA_Exception(origin, "Invalid program namespace: %s", namespace);
		}
		if(TSA_SyntaxPatterns.isValidNamespace(namespace)){
			this.simple_name = simple_name;
		}
		else{
			throw new TSA_Exception(origin, "Invalid simple program name: %s", simple_name);
		}
		//TODO arbitrary descriptions could spoil the xml of a SQL_ProgramSpec
		this.description = description;
		
		this.origin = origin;
		
		this.temporalStreams = new FixedArrayNavigableMap<String, TSA_TemporalStream>(new Adapter<TSA_TemporalStream, String>(){
			@Override
			public String adapt(final TSA_TemporalStream source) {
				return source.full_name;
			}
		}, 
		temporalStreams);
		if(this.temporalStreams.size() < temporalStreams.size()){
			exceptions.add(new TSA_Exception(origin, "Dublicate definition for at least one temporal stream name."));
		}
		

		this.queries = queries;
		for(TSA_Query query: this.queries){
			int index = this.temporalStreams.keys().indexOf(query.output_full_name);
			if(index >= 0){
				if(!this.temporalStreams.values().get(index).schema.equals(query.output_schema)){
					exceptions.add(new TSA_Exception(query.origin, "The schema specification for the output stream of the query differs from the direct schema specification of that stream in the program. Affected temporal stream: " + query.output_full_name));
				}
			}
			else{
				exceptions.add(new TSA_Exception(query.origin, "Query requires undefined temporal stream with name: " + query.output_full_name));
			}
			for(TSA_TemporalStreamExpression input: query.expression.inputs()){
				index = this.temporalStreams.keys().indexOf(input.full_name);
				if(index >= 0){
					if(!TSA_TemporalStreamExpression.adoptSchema(this.temporalStreams.values().get(index).schema, input.full_name, input.origin).equals(input.schema)){
						exceptions.add(new TSA_Exception(query.origin, "The schema specification for an input stream of the query differs from the direct schema specification of that stream in the program. Affected temporal stream: " + input.full_name));
					}
				}
				else{
					exceptions.add(new TSA_Exception(query.origin, "Query requires undefined temporal stream with name: " + input.full_name));
				}
			}
		}
		
		if(exceptions.size() == 1){
			throw exceptions.first();
		}
		else if(exceptions.size() > 1){
			throw new TSA_ComprehensiveException(exceptions);
		}
	}
	
	public boolean isMonotonic(){
		for(TSA_Query query: this.queries){
			if(!query.isMonotonic()){
				return false;
			}
		}
		return true;
	}
	
	public boolean isValid(){
		try{
			this.check();
			//TODO maybe more efficient implementation as isValid does not need to collect exceptions
			return true;
		}
		catch(TSA_Exception e){
			return false;
		}
	}
	
	public void check() throws TSA_Exception{
		ArrayList<TSA_Exception> exceptions = new ArrayList<TSA_Exception>();
		for(TSA_Query query: this.queries){
			try{
				//FIXME check whether query refers to defined streams 
				query.check();
			}
			catch(TSA_Exception e){
				exceptions.add(e);
			}
		}
		//TODO no cyclic dependencies
		if(exceptions.size() == 1){
			throw exceptions.first();
		}
		else if(exceptions.size() > 1){
			throw new TSA_ComprehensiveException(exceptions);
		}
	}
	
	public void writeTo(final StringBuilder sb, final int indentDepth){
		indentTab(sb, indentDepth);
		if(this.namespace != ""){
			sb.append("NAMESPACE ");
			sb.append(this.namespace);
			sb.append("\n");
			sb.append("\n");
			indentTab(sb, indentDepth);
		}
		
		//FIXME description
		
		sb.append("PROGRAM ");
		sb.append(this.simple_name);
		sb.append("\n\n");
		//FIXME
		for(TSA_TemporalStream temporalStream: this.temporalStreams.values()){
			temporalStream.writeTo(sb, indentDepth + 1);
			sb.append("\n\n");
		}
		
		for(TSA_Query query: this.queries){
			query.writeTo(sb, indentDepth + 1);
			sb.append("\n\n");
		}
	}
	
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}

}
