package de.lmu.ifi.pms.cep.tsa.time.constraints;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.BinaryComparisonOperator;

import de.lmu.ifi.pms.java.math.Math;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;

public class SimpleTemporalFormula extends AtomicTemporalRelationFormula{
	
	public final long constant;
	public final String v1;
	public final boolean isAttribute1;
	public final String v2;
	public final boolean isAttribute2;
	public final BinaryComparisonOperator op;

	public final boolean minusEpsilon;
	
	public SimpleTemporalFormula(final BasicTypeID type, final String v1, final BinaryComparisonOperator op, final String v2){
		this(type, v1, op, v2, false);
	}
	
	public SimpleTemporalFormula(final BasicTypeID type, final String v1, final BinaryComparisonOperator op, final String v2, final long constant){
		this(type, v1, false, op, v2, false, constant, false);
	}
	
	public SimpleTemporalFormula(final BasicTypeID type, final String v1, final boolean isAttribute1, final BinaryComparisonOperator op, final String v2, final boolean isAttribute2, final long constant){
		this(type, v1, isAttribute1, op, v2, isAttribute2, constant, false);
	}
	
	/**
	 * This represents v op w +- c
	 * 
	 * @param v1 TemporalVariable 1
	 * @param op Operator chosen from Enum BinaryComparisonOperator
	 * @param v2 TemporalVariable 2
	 * @param pm Plus or Minus
	 * @param constant if constant equals zero we have the special case v op w time in nanosek
	 * 
	 */
	protected SimpleTemporalFormula(final BasicTypeID type, final String v1, final boolean isAttribute1, final BinaryComparisonOperator op, final String v2, final boolean isAttribute2, final long constant, final boolean minusEpsilon){
		super(type);
		this.v1 = v1;
		this.isAttribute1 = isAttribute1;
		this.v2 = v2;
		this.isAttribute2 = isAttribute2;
		this.op = op;
		this.constant = constant;
		if(constant < Math.MinLongValue){
			throw new IllegalArgumentException("Constant is too small.");
		}
		this.minusEpsilon = minusEpsilon;
		if(!(this.op == BinaryComparisonOperator.LESS | this.op == BinaryComparisonOperator.LESS_EQUAL ) && this.minusEpsilon){
			throw new IllegalArgumentException(this.op + " is not allowed in combination with minus epsilon.");
		}
	}
	
	protected SimpleTemporalFormula(final BasicTypeID type, final String v1, final boolean isAttribute1, final BinaryComparisonOperator op, final String v2, final boolean isAttribute2, final boolean minusEpsilon){
		this(type, v1, isAttribute1, op, v2, isAttribute2, 0, minusEpsilon);
	}
	
	public SimpleTemporalFormula(final BasicTypeID type, final String v1, final boolean isAttribute1, final BinaryComparisonOperator op, final String v2, final boolean isAttribute2){
		this(type, v1, isAttribute1, op, v2, isAttribute2, false);
	}
	
	/**
	 * This represents v op w
	 * 
	 * @param v1 TemporalVariable 1
	 * @param op Operator chosen from Enum BinaryComparisonOperator
	 * @param v2 TemporalVariable 2
	 */
	protected SimpleTemporalFormula(final BasicTypeID type, final String v1, final BinaryComparisonOperator op, final String v2, final boolean minusEpsilon){
		this(type, v1, false, op, v2, false, 0, minusEpsilon);
	}
	
	public boolean minusEpsilon() {
		return this.minusEpsilon;
	}
	
	@Override
	protected void VariableNames(final ArrayList<String> variables) {
		if(!this.isAttribute1){
			variables.add(this.v1);
		}
		if(!this.isAttribute2){
			variables.add(this.v2);
		}
	}
	
	@Override
	protected void AttributeNames(final ArrayList<String> attributes) {
		if(this.isAttribute1){
			attributes.add(this.v1);
		}
		if(this.isAttribute2){
			attributes.add(this.v2);
		}
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append((this.isAttribute1 ? "attr " : "var "));
		sb.append(this.v1);		
		sb.append(" ");
		sb.append(this.op.toString());
		sb.append(" ");
		sb.append((this.isAttribute2 ? "attr " : "var "));
		sb.append(this.v2);
		if(this.constant != 0){
			sb.append(" + ");
			sb.append(this.constant);	
		}
		if(this.minusEpsilon){
			sb.append("-e");
		}
	}
	
	@Override
	public TemporalRelationFormula flaten() {
		return this;
	}
	
	@Override
	protected TemporalRelationFormula simpleForm(final boolean positivePolarity){
		if(positivePolarity){
			//Eq :   v = w + c   ->   (v <= w + c) AND (w <= v + (-c))
			//Less : v < w + c   ->    v <= w + (c-epsilon)		
			switch(this.op){
				case LESS : return new SimpleTemporalFormula(this.type, this.v1, this.isAttribute1, BinaryComparisonOperator.LESS_EQUAL, this.v2, this.isAttribute2, this.constant, true);
				case LESS_EQUAL : return this;
				case EQUAL : assert !this.minusEpsilon; return new ConjunctiveTemporalRelationFormula(new SimpleTemporalFormula(this.type, this.v1, this.isAttribute1, BinaryComparisonOperator.LESS_EQUAL,this.v2, this.isAttribute2,this.constant,false),new SimpleTemporalFormula(this.type, this.v2, this.isAttribute2, BinaryComparisonOperator.LESS_EQUAL, this.v1, this.isAttribute1, -this.constant,false));
				case GREATER: assert !this.minusEpsilon; return new SimpleTemporalFormula(this.type, this.v2, this.isAttribute2, BinaryComparisonOperator.LESS_EQUAL, this.v1, this.isAttribute1, -this.constant, true);
				case GREATER_EQUAL: assert !this.minusEpsilon; return new SimpleTemporalFormula(this.type, this.v2, this.isAttribute2, BinaryComparisonOperator.LESS_EQUAL, this.v1, this.isAttribute1, -this.constant, false);
				case NOT_EQUAL: assert !this.minusEpsilon; return new DisjunctiveTemporalRelationFormula(new SimpleTemporalFormula(this.type, this.v1, this.isAttribute1,BinaryComparisonOperator.LESS_EQUAL,this.v2, this.isAttribute2,this.constant,true),new SimpleTemporalFormula(this.type, this.v2, this.isAttribute2, BinaryComparisonOperator.LESS_EQUAL, this.v1, this.isAttribute1, -this.constant,true));
				default: throw new Error("Uncovered case: " + this.op);
			}
		}
		else{
			return this.negatedForm().simpleForm(true);
		}
	}
	

	public TemporalRelationFormula negatedForm(){
		// v <= w + c wird zu w < v + (-c)
		// v < w + c wird zu w <= v + (-c) 
		// v = w + c wird zu v < w + c OR w < v + (-c)
		
		switch(this.op){
			case LESS         : return new SimpleTemporalFormula(this.type, this.v2, this.isAttribute2, BinaryComparisonOperator.LESS_EQUAL, this.v1, this.isAttribute1, -this.constant, false); 
			case LESS_EQUAL   : return this.minusEpsilon ? new SimpleTemporalFormula(this.type, this.v2, this.isAttribute2, BinaryComparisonOperator.LESS_EQUAL, this.v1, this.isAttribute1, -this.constant, false) : new SimpleTemporalFormula(this.type, this.v2, this.isAttribute2, BinaryComparisonOperator.LESS, this.v1, this.isAttribute1, -this.constant, false);
			case EQUAL        : assert !this.minusEpsilon; return new DisjunctiveTemporalRelationFormula(new SimpleTemporalFormula(this.type, this.v1, this.isAttribute1, BinaryComparisonOperator.LESS, this.v2, this.isAttribute2,this.constant,false),new SimpleTemporalFormula(this.type, this.v2, this.isAttribute2, BinaryComparisonOperator.LESS, this.v1, this.isAttribute1,-this.constant,false));
			case GREATER      : assert !this.minusEpsilon; return new SimpleTemporalFormula(this.type, this.v1, this.isAttribute1, BinaryComparisonOperator.LESS_EQUAL, this.v2, this.isAttribute2, this.constant, false);
			case GREATER_EQUAL: assert !this.minusEpsilon; return new SimpleTemporalFormula(this.type, this.v1, this.isAttribute1, BinaryComparisonOperator.LESS, this.v2, this.isAttribute2, this.constant, false);
			case NOT_EQUAL    : assert !this.minusEpsilon; return new SimpleTemporalFormula(this.type, this.v1, this.isAttribute1, BinaryComparisonOperator.EQUAL, this.v2, this.isAttribute2, this.constant, false);
			default           : throw new Error("Uncovered case: " + this.op);
		}

	}
	
	protected AtomicTerm<BasicType> arg1(){
		if(this.isAttribute1){
			return BasicTypeAttribute.create(this.v1, this.type);
		}
		else{
			return BasicTypeVariable.create(this.v1, this.type);
		}
	}
	
	
	protected AtomicTerm<BasicType> arg2(){
		if(this.isAttribute2){
			return BasicTypeAttribute.create(this.v2, this.type);
		}
		else{
			return BasicTypeVariable.create(this.v2, this.type);
		}
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		AtomicTerm<BasicType> arg1 = this.arg1();
		if(filter.accepts(arg1)){
			terms.add(arg1);
		}
		AtomicTerm<BasicType> arg2 = this.arg2();
		if(filter.accepts(arg2)){
			terms.add(arg2);
		}
	}
	
	@Override
	public TemporalRelationFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		Term<?> t1 = replacement.adapt(this.arg1());
		Term<?> t2 = replacement.adapt(this.arg2());
		String v1;
		boolean isAttribute1;
		BasicTypeID type1;
		String v2;
		boolean isAttribute2;
		BasicTypeID type2;
		if(t1 instanceof BasicTypeAttribute){
			v1 = ((BasicTypeAttribute) t1).name;
			type1 = ((BasicTypeAttribute) t1).type.ID;
			isAttribute1 = true;
		}
		else if(t1 instanceof BasicTypeVariable){
			v1 = ((BasicTypeVariable) t1).name;
			type1 = ((BasicTypeVariable) t1).type.ID;
			isAttribute1 = false;
		}
		else{
			throw new IllegalArgumentException();
			//TODO better exception
		}
		if(t2 instanceof BasicTypeAttribute){
			v2 = ((BasicTypeAttribute) t2).name;
			type2 = ((BasicTypeAttribute) t2).type.ID;
			isAttribute2 = true;
		}
		else if(t2 instanceof BasicTypeVariable){
			v2 = ((BasicTypeVariable) t2).name;
			type2 = ((BasicTypeVariable) t2).type.ID;
			isAttribute2 = false;
		}
		else{
			throw new IllegalArgumentException();
			//TODO better exception
		}
		if(v1.equals(this.v1) & v2.equals(this.v2) & type1 == this.type & type2 == this.type & isAttribute1 == this.isAttribute1 & isAttribute2 == this.isAttribute2){
			return this;
		}
		else{
			if(type1 == type2){
				return new SimpleTemporalFormula(type1, v1, isAttribute1, this.op, v2, isAttribute2, this.constant, this.minusEpsilon);
			}
			else{
				throw new IllegalArgumentException();
				//TODO better exception				
			}
		}
				
	}
	
	@Override
	public TemporalRelationFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		return this.replaceAtomicTerm(replacement);
	}
	
}
